#!/usr/bin/perl -w

#Gianduia, DGAS job monitor and Usage Records collector.
use strict;
use POSIX;
use Text::ParseWords;

my $DEF_LDIF_VALIDITY = 86400; # assume the GLUE attributes taken from the LDIF
                               # didn't change within the last day.

my $sigset = POSIX::SigSet ->new();
my $actionHUP=POSIX::SigAction->new("sigHUP_handler",$sigset,&POSIX::SA_NODEFER);
my $actionInt=POSIX::SigAction->new("sigINT_handler",$sigset,&POSIX::SA_NODEFER);
POSIX::sigaction(&POSIX::SIGHUP, $actionHUP);
POSIX::sigaction(&POSIX::SIGINT, $actionInt);
POSIX::sigaction(&POSIX::SIGTERM, $actionInt);
my $configFilePath ="/opt/glite/etc/dgas_gianduia.conf"; 
my $lastId = "";
my %configValues = (
		    lsfAcctLogDir     =>  "",
		    lockFileName      => "/opt/glite/var/dgasOOB.lock",
		    bufferFileName      => "/opt/glite/var/dgasOOBuffer",
		    mainPollInterval  => "5",
		    queuePollInterval =>"25",
		    timeInterval      =>"5",
		    jobPerTimeInterval =>"10",
		    siteName =>"",
		    localHostNameAsCEHostName => "no",

		    keyList => "GlueHostBenchmarkSF00,GlueHostBenchmarkSI00",
		    ldifDefaultFiles => "",
		    glueLdifFile => "",
		    );

my %hash = ();

#Parse configuration file
my $startJob = "";
my $lastJob = "";
if(exists $ARGV[0]) 
{
    $configFilePath = $ARGV[0];
}
if(exists $ARGV[1])
{
    #start analising from $lastId and go backward instead
    #of starting from end of file.
    $lastId = $ARGV[1];	
    if(exists $ARGV[2])
    {
	$startJob = $ARGV[2];
    }
}
&parseConf($configFilePath); 

## Global variable initialization
my $UR;
my $list;

my $lsfAcctLogDir    = $configValues{lsfAcctLogDir};
my $timeInterval     = $configValues{timeInterval};
my $jobPerTimeInterval     = $configValues{jobPerTimeInterval};

my $lockFileName = $configValues{lockFileName};

my $keyList = $configValues{keyList};
my $ldifDefaultFiles = $configValues{ldifDefaultFiles};
my $glueLdifFile = $configValues{glueLdifFile};
my %glueAttributes = ();
my $ldifModTimestamp = 0;

my $timeZone = `date +%z`;
chomp($timeZone);
my $hostName = `hostname -s`;
chomp($hostName);
my $domainName = `hostname -d`;
chomp($domainName);
my $keepGoing = 1;
my $start1 = time;   ##  time init
my $start2 = time;   ##  time init
my @garbageQueue = ();

#put lock
if ( &putLock($lockFileName) != 0 )
{
    die "Fatal Error: Couldn't open lock file! in $lockFileName\n";
}
else
{
	print "Lock file succesfully created.\n";
}

#This script must be exexuted in the cron.
if ( $startJob eq "" )
{
	&readBuffer($configValues{bufferFileName}, $startJob);
}
print "First job analysed: $startJob\n";

# get GLUE attributes from LDIF files.
$ldifModTimestamp = &getGLUEAttributesFromLdif();

if (&checkLsfLog($startJob,$lastJob) == 0)
{
    print "Last job analysed: $lastJob\n";
    if ($lastJob ne "" )
    {
	&putBuffer($configValues{bufferFileName}, $lastJob);
    }
}
else {
    print "Error while processing LRMS logs backwards! Not writing new buffer file!\n";
}


print "Exiting...\n";
if ( &delLock($lockFileName) != 0 )
{	
    print "Error removing lock file.\n";
}
else
{
    print "Lock file removed.\n";
}
print "Exit.\n";


###-----------------------END-----------------------------------------###

sub parseConf 
{
    my $fconf = $_[0];
    open(FILE, "$fconf") || die "Error: Cannot open configuration file: $fconf\n";
    while(<FILE>)
    {
	if(/^lsfAcctLogDir\s*=\s*\"(.*)\"$/){$configValues{lsfAcctLogDir}=$1;}
	if(/^oobLockFileName\s*=\s*\"(.*)\"$/){$configValues{lockFileName}=$1;}
	if(/^bufferFileName\s*=\s*\"(.*)\"$/){$configValues{bufferFileName}=$1;}
	if(/^mainPollInterval\s*=\s*\"(.*)\"$/){$configValues{mainPollInterval}=$1;}
	if(/^queuePollInterval\s*=\s*\"(.*)\"$/){$configValues{queuePollInterval}=$1;}
	if(/^siteName\s*=\s*\"(.*)\"$/){$configValues{siteName}=$1;}
	if(/^localHostNameAsCEHostName\s*=\s*\"(.*)\"$/){$configValues{localHostNameAsCEHostName}=$1;}

	if(/^keyList\s*=\s*\"(.*)\"$/){$configValues{keyList}=$1;}
	if(/^ldifDefaultFiles\s*=\s*\"(.*)\"$/){$configValues{ldifDefaultFiles}=$1;}
	if(/^glueLdifFile\s*=\s*\"(.*)\"$/){$configValues{glueLdifFile}=$1;}
    }
    close(FILE);
}

sub checkLsfLog#FIXME
{
        my $startJob = $_[0];
	
        $lsfAcctLogDir= $configValues{lsfAcctLogDir};
        open ( LOGFILE, "/usr/bin/tac $lsfAcctLogDir/lsb.acct |" ) or return 1;
	if ( $lastId eq "" )
	{
		print "Parsing file from end to $startJob in $lsfAcctLogDir/lsb.acct\n";
	}
	else
	{
		 print "Parsing file from $lastId to $startJob in $lsfAcctLogDir/lsb.acct\n";
	}
	my $firstJobId = 1;
        while ( <LOGFILE> )
        {
		if ( $keepGoing == 0 )
		{
			print "Exiting from LSF loop...\n";
			close LOGFILE;
			return 0;
		}
                my @ARRAY = split(" " , $_ );
                my $lsfVersion = $ARRAY[1];
                $lsfVersion =~ s/\"//g;
                if ( $lsfVersion eq "4.1" ||
                        $lsfVersion eq "5.0" ||
                        $lsfVersion eq "5.1" ||
                        $lsfVersion eq "6.0" ||
                        $lsfVersion eq "6.1" )
                {
                        my $currentJobId = $ARRAY[3];
			if ( ($currentJobId ne $lastId) && ($lastId ne "") )
			{
				next;
			}
			else
			{
				$lastId = ""; ##so next check shall fail.
			}
                        if ( ( $startJob eq $currentJobId ) && ($startJob ne "") )
                        {
                                #Got the right line!
                                print "Found $startJob in LOG!\n";
				close LOGFILE;
                                return 0;
                        }
			else
			{
				#this record should be accounted
				if ( $firstJobId == 1 )
				{
					$_[1] = $currentJobId;
					$firstJobId = 0;
				}
				print "$currentJobId:";
				&parseUR_LSF($_);
				print "processing $hash{jobName}\n";
				if(&callAtmClient() == 0)
				{
				    print "UR successfully forwarded to HLR.\n";
				}
				else
				{
				    print "Error! Could not push UR to HLR! Stopping iteration!\n";
				    return 1;
				}
			}
                }
        }
        close LOGFILE;
        return 0;
}

sub parseUR_LSF
{

        %hash = ();

        my $URString = $_[0];
        my @new = quotewords(" ", 0, $URString);
        my $shift1 = $new[22];
        my $shift2 = $new[23+$shift1];
        #$hash{server}=$new[16].".".$domainName;
	# attention: LSF log might already contain a complete hostname
	# (with domain name) ...
	# cross-check, instead of just adding the CE's domain name:
	$hash{server}=$new[16];
	if ($hash{server} =~ /^([^\.]*)\.(.*)$/)
	{
	    # if we found a domain name in the LSF log:
	    if ( ($2 ne $domainName) && ($domainName ne "") )
	    {
		# replace by CE's domain name if different:
		$hash{server} = "$1.$domainName";
	    }
	}
	elsif ($domainName ne "")
	{
	    # if no domain name in LSF log: add the CE's domain name
	    $hash{server} .= ".$domainName";
	}

        $hash{queue}=$new[12];
        $hash{user}=$new[11];
        $hash{lrmsId}=$new[3];
        $hash{processors}=$new[6];
	if ($new[10] == 0)
	{   # happens if job was cancelled before executing?
	    $hash{walltime}=0;
	}
	else
	{
	    $hash{walltime}=$new[2]-$new[10];
	}
	if ($new[28+$shift2] == -1)
	{   # indicates that the value is not available!
	    $new[28+$shift2]=0;
	}
	if ($new[29+$shift2] == -1)
	{   # indicates that the value is not available!
	    $new[29+$shift2]=0;
	}
        $hash{cput}=int($new[28+$shift2])+int($new[29+$shift2]);
        $hash{mem}=$new[54+$shift2]."k";
        $hash{vmem}=$new[55+$shift2]."k";
        $hash{start}=$new[10];
        $hash{end}=$new[2];
        $hash{jobName}=$new[26+$shift2];
        $hash{exitStatus}=$new[49+$shift2];
	&detectVOfromuid($hash{user});
}

sub detectVOfromuid
{
	my $localuid=$_[0];
	if($localuid =~ /^(.*)sgm$/ || $localuid =~ /^(.*)prd$/ || $localuid =~ /^(.*)\d{3}$/)
	{
		$hash{userVO}=$1;
	}
	else
	{
		$hash{userVO}="";
	}
	
}

sub callAtmClient 
{
    my $now_string = localtime;
    if ( $keepGoing == 0 )
    {
        print "$now_string: Got termination signal..\n";
        return -2;
    }


    if ($configValues{localHostNameAsCEHostName} eq "yes") {
	$hash{server} = $hostName.".".$domainName;
    }

    my $cmd = "";
    my $status = -1; 
    my $resGridId="$hash{server}:$hash{queue}";
    $cmd = "/opt/glite/libexec/glite_dgas_atmClient";
    $cmd .= " --jobid $hash{server}:$hash{lrmsId}_$hash{start}";
    $cmd .= " --resgridid $resGridId";
    $cmd .= " CPU_TIME=$hash{cput} WALL_TIME=$hash{walltime} PMEM=$hash{mem} VMEM=$hash{vmem}";
    #$cmd .= " \"QUEUE=$hash{queue}\" \"USER=$hash{user}\" \"LRMSID=$hash{lrmsId}\"";
    if (exists($hash{queue})) {
	$cmd .= " \"QUEUE=$hash{queue}\"";
    }
    if (exists($hash{user})) {
	$cmd .= " \"USER=$hash{user}\"";
    }
    if (exists($hash{lrmsId})) {
	$cmd .= " \"LRMSID=$hash{lrmsId}\"";
    }
    #$cmd .= " \"PROCESSORS=$hash{processors}\" \"URCREATION=$now_string\"";
    if (exists($hash{processors})) {
	$cmd .= " \"PROCESSORS=$hash{processors}\"";
    }
    $cmd .= " \"URCREATION=$now_string\"";
    #$cmd .= " \"group=$hash{group}\" \"jobName=$hash{jobName}\"";
    if (exists($hash{group})) {
	$cmd .= " \"group=$hash{group}\"";
    }
    if (exists($hash{jobName})) {
	$cmd .= " \"jobName=$hash{jobName}\"";
    }
    #$cmd .= " \"start=$hash{start}\" \"end=$hash{end}\"";
    if (exists($hash{start})) {
	$cmd .= " \"start=$hash{start}\"";
    }
    if (exists($hash{end})) {
	$cmd .= " \"end=$hash{end}\"";
    }
    #$cmd .= " \"ctime=$hash{ctime}\" \"qtime=$hash{qtime}\"";
    if (exists($hash{ctime})) {
	$cmd .= " \"ctime=$hash{ctime}\"";
    }
    if (exists($hash{qtime})) {
	$cmd .= " \"qtime=$hash{qtime}\"";
    }
    #$cmd .= " \"etime=$hash{etime}\" \"exitStatus=$hash{exitStatus}\"";
    if (exists($hash{etime})) {
	$cmd .= " \"etime=$hash{etime}\"";
    }
    if (exists($hash{exitStatus})) {
	$cmd .= " \"exitStatus=$hash{exitStatus}\"";
    }

    # we don't have this from the LRMS log, but from the LDIF file(s):
    ##$cmd .= " \"si2k=$hash{si2k}\" \"sf2k=$hash{sf2k}\"";
    #if (exists($hash{si2k})) {
	#$cmd .= " \"si2k=$hash{si2k}\"";
    #}
    #if (exists($hash{sf2k})) {
	#$cmd .= " \"sf2k=$hash{sf2k}\"";
    #}

    # If we found an LDIF file ($ldifModTimestamp != 0)
    # and this job is more recent than the file, then we trust the
    # SpecInt and SpecFloat values taken from the file:
    # (IMPORTANT: for now we DO NOT check the age of the job and ALWAYS
    #  accept the LDIF attributes!)
    if (exists($hash{start}) && ($ldifModTimestamp != 0) 
	#&& ($hash{start} >= $ldifModTimestamp) 
	) {
	if (exists($glueAttributes{"GlueHostApplicationSoftwareRunTimeEnvironment: SI00MeanPerCPU"})
	    && $glueAttributes{"GlueHostApplicationSoftwareRunTimeEnvironment: SI00MeanPerCPU"} ne "") {
	    my $si2k = $glueAttributes{"GlueHostApplicationSoftwareRunTimeEnvironment: SI00MeanPerCPU"};
	    $cmd .= " \"si2k=$si2k\"";
	    print "(note: Taking mean SI00 value per CPU, instead of the normal benchmark)\n";
	}
	elsif (exists($glueAttributes{GlueHostBenchmarkSI00})
	       && $glueAttributes{GlueHostBenchmarkSI00} ne "") {
	    $cmd .= " \"si2k=$glueAttributes{GlueHostBenchmarkSI00}\"";
	}

	if (exists($glueAttributes{"GlueHostApplicationSoftwareRunTimeEnvironment: SF00MeanPerCPU"})
	    && $glueAttributes{"GlueHostApplicationSoftwareRunTimeEnvironment: SF00MeanPerCPU"} ne "") {
	    my $sf2k = $glueAttributes{"GlueHostApplicationSoftwareRunTimeEnvironment: SF00MeanPerCPU"};
	    $cmd .= " \"sf2k=$sf2k\"";
	    print "(note: Taking mean SF00 value per CPU, instead of the normal benchmark)\n";
	}
	elsif (exists($glueAttributes{GlueHostBenchmarkSF00})
	       && $glueAttributes{GlueHostBenchmarkSF00} ne "") {
	    $cmd .= " \"sf2k=$glueAttributes{GlueHostBenchmarkSF00}\"";
	}
    }

    $cmd .= " \"tz=$timeZone\" \"accountingProcedure=outOfBand\"";
    #$cmd .= " \"fqan=/$hash{userVO}/Role=NULL/Capability=NULL\""; #!!!!!!
    if (exists($hash{userVO})) {
	$cmd .= " \"fqan=/$hash{userVO}/Role=NULL/Capability=NULL\"";
    }
    $cmd .= " \"ceHostName=$hash{server}\""; 
    if (exists($hash{execHost})) {
	$cmd .= " \"execHost=$hash{execHost}\"";
    }
    if ( $configValues{siteName} ne "" )
    { 
            $cmd .= " \"SiteName=$configValues{siteName}\"";
    }
    $cmd .= " --usrcert \"$hash{user}:$hash{server}\"";
    $cmd .= " -3";


    $status = system("$cmd");
    print "$now_string: Running: $cmd\n";
    if ($? == -1) {
           print "failed to execute: $!\n";
    }
    elsif ($? & 127) {
           $status = ($? & 127);
    }
    else {
           $status = $? >> 8;
    }
    print "EXITSTATUS=$status\n";

    if ( $status == 64 || ($status>=67 && $status<=69) ) {
	# from hlr_prot_errcode.h:
        #define ATM_E_TRANS 		"64"
        #define ATM_E_TRANS_A 		"67"
        #define ATM_E_TRANS_B 		"68"
        #define ATM_E_TRANS_C 		"69"
	print "Attention: The record cannot be inserted into the HLRs transaction queue. Terminating, job will be retried next time (if this conditions persists, there might be problems on the HLR (database down, record with same jobID already in queue cannot be processed, etc).\n";
    }

    if ( $status == 65 || ($status>=70 && $status<=73) ) {
	# from hlr_prot_errcode.h:
        #define ATM_E_DUPLICATED 	"65"
        #define ATM_E_DUPLICATED_A 	"70"
        #define ATM_E_DUPLICATED_B 	"71"
        #define ATM_E_DUPLICATED_C 	"72"
        #define ATM_E_DUPLICATED_D 	"73"
	print "The record is already accounted on the HLR.\n";
	return 0;
    }

    print "\n";

    return $status;
}


##-------> lock  subroutines <---------##

sub putLock
{
    my $lockName = $_[0];
    open(IN,  "< $lockName") && return 1;
    close(IN);
    open(OUT, "> $lockName") || return 2;
    print OUT  $$;    ## writes pid   
    close(OUT);
    return 0;
}

sub delLock
{
    my $lockName = $_[0];
    open(IN,  "< $lockName") || return 1;
    close(IN);
    my $status = system("rm -f $lockName");
    return $status;
}

##-------> sig handlers subroutines <---------##

sub putBuffer
{
    my $buffName = $_[0];
    if ($keepGoing == 1) {
	# this is done only if no SIGINT was received!
	open(OUT, "> $buffName") || return 2;
	print OUT  $_[1];    ## writes lastId
	close(OUT);
    }
    return 0;
}

sub readBuffer
{
	my $buffname = $_[0];
	open(IN, "< $buffname") || return 2;
	while ( <IN> )
	{
		my $sJob = $_;
		chomp($sJob);
		$_[1] = $sJob;
	}
	close(IN);
	return 0;
}

sub sigHUP_handler {
        print "got SIGHUP\n";
	$keepGoing = 1;
}

sub sigINT_handler {
        print "got SIGINT\n";
        $keepGoing = 0;
}



## ------ get GLUE attributes From LDIF file --------- ##
sub getGLUEAttributesFromLdif {
    my $modTStamp = 0; # returns 0 if no valid attributes found!
                       # otherwise returns the timestamp of the last
                       # modification of the file!

    # to be stored in %glueAttributes
    %glueAttributes = (); # first empty everything

    my @ldifFiles = split(/,/, $ldifDefaultFiles);
    unshift(@ldifFiles, $glueLdifFile); # first to try!

    my @keys = split(/,/, $keyList);

    if (!@keys) {
	print "".localtime().": Warning: No GLUE attributes will be added to usage records (reason: no keyList in configuration file)!\n";
	return 0; # no keys -> no benchmarks
    }

    while (@ldifFiles) {
	my $file = shift(@ldifFiles);
	print "".localtime().": Trying to get GLUE benchmarks from LDIF file: $file\n";

	if (!open(GLUEFILE, "< $file")) {
	    print "Warning: could not open the LDIF file ... skipping!\n";
	    next;
	}

	my $foundSomething = 0;
	my $line;
	while ($line = <GLUEFILE>) {
	    my $key;
	    foreach $key (@keys) {
		if ( ($line =~ /^$key:\s?(.*)$/ )
		     || ($line =~ /^$key=(.*)$/ ) 
		     || ($line =~ /^${key}_(.*)$/ ) ) {
		    # accept stuff like "GlueHostBenchmarkSI00: 955" and
		    # "GlueHostApplicationSoftwareRunTimeEnvironment: SI00MeanPerCPU=955"
		    # "GlueHostApplicationSoftwareRunTimeEnvironment: SI00MeanPerCPU_955"
		    $glueAttributes{$key} = $1;
		    print "found: $key=$1; ";
		    $foundSomething = 1;
		}
	    }
	}

	close(GLUEFILE);

	if ($foundSomething) {
	    print "\n";
	    # get timestamp of file:
	    $modTStamp = (stat($file))[9]; #mtime in sec. after epoch

	    last; # don't check the other files (if any)!
	}

    }

    my $thisTStamp = time();

    if ($modTStamp == 0) {
	print "".localtime().": Warning: No GLUE attributes will be added to usage records (reason: no valid entries in LDIF file(s))!\n";
	return 0;
    }
    else {
	if($modTStamp > $thisTStamp-$DEF_LDIF_VALIDITY) {
	    # accept it for at least a day, even if more recently modified
	    $modTStamp = $thisTStamp-$DEF_LDIF_VALIDITY;
	}
	# we disabled the controll above!
	#print "GLUE attributes will be accepted for all jobs with LRMS creation time after $modTStamp\n";
    }



    return $modTStamp;
}
