#!/usr/bin/perl -w

#Gianduia, DGAS job monitor and Usage Records collector.
use strict;
use POSIX;
use Text::ParseWords;

my $DEF_LDIF_VALIDITY = 86400; # assume the GLUE attributes taken from the LDIF
                               # didn't change within the last day.

my $sigset = POSIX::SigSet ->new();
my $actionHUP=POSIX::SigAction->new("sigHUP_handler",$sigset,&POSIX::SA_NODEFER);
my $actionInt=POSIX::SigAction->new("sigINT_handler",$sigset,&POSIX::SA_NODEFER);
POSIX::sigaction(&POSIX::SIGHUP, $actionHUP);
POSIX::sigaction(&POSIX::SIGINT, $actionInt);
POSIX::sigaction(&POSIX::SIGTERM, $actionInt);
my $configFilePath ="/opt/glite/etc/dgas_gianduia.conf"; 
my %configValues = (
		    pbsAcctLogDir     =>  "",
		    lockFileName      => "/opt/glite/var/dgasOOB.lock",
		    bufferFileName      => "/opt/glite/var/dgasOOBuffer",
		    mainPollInterval  => "5",
		    queuePollInterval =>"25",
		    timeInterval      =>"5",
		    jobPerTimeInterval =>"10",
		    siteName => "",
		    localHostNameAsCEHostName => "no",

		    keyList => "GlueHostBenchmarkSF00,GlueHostBenchmarkSI00",
		    ldifDefaultFiles => "",
		    glueLdifFile => "",
		    );

my %hash = ();

#Parse configuration file
if(exists $ARGV[0]) 
{
    $configFilePath = $ARGV[0];
}	
&parseConf($configFilePath); 

## Global variable initialization
my $UR;
my $list;

my $pbsAcctLogDir    = $configValues{pbsAcctLogDir};
my $timeInterval     = $configValues{timeInterval};
my $jobPerTimeInterval     = $configValues{jobPerTimeInterval};

my $lockFileName = $configValues{lockFileName};

my $keyList = $configValues{keyList};
my $ldifDefaultFiles = $configValues{ldifDefaultFiles};
my $glueLdifFile = $configValues{glueLdifFile};
my %glueAttributes = ();
my $ldifModTimestamp = 0;

my $T2 = $configValues{queuePollInterval};
my $T1 = $configValues{mainPollInterval};
my $garbageCleanInterval = 10*$T2;

my $timeZone = `date +%z`;
chomp($timeZone);
my $hostName = `hostname -s`;
chomp($hostName);
my $domainName = `hostname -d`;
chomp($domainName);
my $keepGoing = 1;
my $start1 = time;   ##  time init
my $start2 = time;   ##  time init
my @garbageQueue = ();

#put lock
if ( &putLock($lockFileName) != 0 )
{
    die "Fatal Error: Couldn't open lock file! in $lockFileName\n";
}
else
{
	print "Lock file succesfully created.\n";
}

#This script must be exexuted in the cron.
my $startJob = "";
my $startFile = "";
my $lastJob = "";
my $lastFile = "";
&readBuffer($configValues{bufferFileName}, $startJob, $startFile);
print "First job to analyse: $startFile:$startJob\n";

# get GLUE attributes from LDIF files.
$ldifModTimestamp = &getGLUEAttributesFromLdif();

if (&checkPbsLog($startJob,$startFile,$lastFile,$lastJob) == 0)
{    
    print "Last job analysed: $lastFile:$lastJob\n";
    if ($lastJob ne "" )
    {
	&putBuffer($configValues{bufferFileName}, $lastJob, $lastFile);
    }
}
else {
    print "Error while processing LRMS logs backwards! Not writing new buffer file!\n";
}


print "Exiting...\n";
if ( &delLock($lockFileName) != 0 )
{	
    print "Error removing lock file.\n";
}
else
{
    print "Lock file removed.\n";
}
print "Exit.\n";


###-----------------------END-----------------------------------------###

sub parseConf 
{
    my $fconf = $_[0];
    open(FILE, "$fconf") || die "Error: Cannot open configuration file: $fconf\n";
    while(<FILE>)
    {
	if(/^pbsAcctLogDir\s*=\s*\"(.*)\"$/){$configValues{pbsAcctLogDir}=$1;}
	if(/^oobLockFileName\s*=\s*\"(.*)\"$/){$configValues{lockFileName}=$1;}
	if(/^bufferFileName\s*=\s*\"(.*)\"$/){$configValues{bufferFileName}=$1;}
	if(/^mainPollInterval\s*=\s*\"(.*)\"$/){$configValues{mainPollInterval}=$1;}
	if(/^queuePollInterval\s*=\s*\"(.*)\"$/){$configValues{queuePollInterval}=$1;}
	if(/^siteName\s*=\s*\"(.*)\"$/){$configValues{siteName}=$1;}
	if(/^localHostNameAsCEHostName\s*=\s*\"(.*)\"$/){$configValues{localHostNameAsCEHostName}=$1;}

	if(/^keyList\s*=\s*\"(.*)\"$/){$configValues{keyList}=$1;}
	if(/^ldifDefaultFiles\s*=\s*\"(.*)\"$/){$configValues{ldifDefaultFiles}=$1;}
	if(/^glueLdifFile\s*=\s*\"(.*)\"$/){$configValues{glueLdifFile}=$1;}
    }
    close(FILE);
}

sub checkPbsLog#FIXME
{
        my $startJob = $_[0]; #jobNumber to start with.
        my $startFile = $_[1]; #jobNumber to start with.
	my $lastFile= $_[2]; #last file analysed on previus run
        $pbsAcctLogDir= $configValues{pbsAcctLogDir};
	my @dirs;
	opendir(DIR,$pbsAcctLogDir ) || die "can't opendir $pbsAcctLogDir: $!";
	while( defined(my $file = readdir(DIR)) )
	{
		push @dirs, $file;
	}
	closedir DIR;
	my $goOn = 1;
	my $firstJobId = 1;
	my $firstFileOpened = 0;
	print "Scanning logdir $pbsAcctLogDir until file:$startFile\n";
	while ( $goOn && $keepGoing && @dirs && !$firstFileOpened )
	{
		my $file = pop @dirs; #going backward
		next if (($file eq ".") || ($file eq ".."));
		if ( $file eq $startFile )
		{
			##this file has already been fully processed
			##so we can break here.
			#print "Found $startFile in dir no need to go further!\n";
			#return 0;
			print "Found $startFile in dir; last file to process!\n";
			$firstFileOpened = 1;
		}
		open ( LOGFILE, "/usr/bin/tac $pbsAcctLogDir/$file |" ) or return 1;
		print "Parsing file from end to $startJob in $pbsAcctLogDir/$file\n";
		while ( <LOGFILE> )
		{
			if ( $keepGoing == 0 )
			{
				print "Exiting from PBS loop...\n";
				close LOGFILE;
				return 0;
			}
			my @ARRAY = split(" " , $_ );
			my @ARRAY2 = split(";" , $ARRAY[1] );
			my $targetJobId = $ARRAY2[2];
			my $lineType = $ARRAY2[1];
			if ( $lineType ne "E" )
			{
				next;
			}
			if ( ( $startJob eq $targetJobId ) && ($startJob ne "") )
			{
				#Got the right line!
				print "Found $targetJobId in LOG!\n";
				close LOGFILE;
				return 0;
			}	
			else
			{
				if ( $firstJobId == 1 )
				{
					$_[2] = $file;
					$_[3] = $targetJobId;
					$firstJobId = 0;
				}
				print "$targetJobId:";
				&parseUR_PBS($_);
				print "processing $hash{jobName}\n";
				if(&callAtmClient() == 0)
				{
				    print "UR successfully forwarded to HLR.\n";
				}
				else
				{
				    print "Error! Could not push UR to HLR! Stopping iteration!\n";
				    return 1;
				}
			}
		}
	}

	return 0;
}

sub parseUR_PBS
{

	%hash = ();

	my $URString = $_[0];
	my @URArray = split ( ' ' , $URString );
	my @tmpArray = split ( ';', $URArray[1] );
	$_ = $tmpArray[3];
	if (/^user=(.*)$/){$hash{user}=$1;};
	$hash{lrmsId}=$tmpArray[2];
	$_ = $tmpArray[2];
	if (/^(\d*)\.(.*)$/){$hash{server}=$2;};
	foreach ( @URArray )
	{
		if ( /^queue=(.*)$/){$hash{queue}=$1;};
		if ( /^resources_used.cput=(.*)$/)
                {
                        $_=$1;
                        $_=~ /(\d\d):(\d\d):(\d\d)$/;
                        $hash{cput}= $3 + $2*60 + $1*3600;
                }
                if ( /^resources_used.walltime=(.*)$/)
                {
                        $_=$1;
                        $_=~ /(\d\d):(\d\d):(\d\d)$/;
                        $hash{walltime}= $3 + $2*60 + $1*3600;
                }
                if ( /^resources_used.vmem=(.*)$/)
                {
                        $_=$1;
                        $_=~ /(\d*[M.k]b)$/;
                        $hash{vmem}     = $1;
                }
                if ( /^resources_used.mem=(.*)$/)
                {
                        $_=$1;
                        $_=~ /(\d*[M.k]b)$/;
                        $hash{mem}     = $1;
                }
                if ( /^Resource_List.neednodes=(.*)$/)
                {
                        $hash{processors} = $1;
                }
		if ( /^group=(.*)$/)
                {
                        $hash{group} = $1;
                }
                if ( /^jobname=(.*)$/)
                {
                        $hash{jobName} = $1;
                }
                if ( /^ctime=(.*)$/)
                {
                        $hash{ctime} = $1;
                }
                if ( /^qtime=(.*)$/)
                {
                        $hash{qtime} = $1;
                }
                if ( /^etime=(.*)$/)
                {
                        $hash{etime} = $1;
                }
                if ( /^start=(.*)$/)
                {
                        $hash{start} = $1;
                }
                if ( /^end=(.*)$/)
                {
                        $hash{end} = $1;
                }
                if ( /^exec_host=(.*)$/)
                {
                        $hash{execHost} = $1;
                }
                if ( /^Exit_status=(.*)$/)
                {
                        $hash{exitStatus} = $1;
                }
	}
	&detectVOfromuid($hash{user});
}

sub detectVOfromuid
{
	my $localuid=$_[0];
	if($localuid =~ /^(.*)sgm$/ || $localuid =~ /^(.*)prd$/ || $localuid =~ /^(.*)\d{3}$/)
        {
                $hash{userVO}=$1;
        }
        else
        {
                $hash{userVO}="";
        }
}

sub callAtmClient 
{
    my $now_string = localtime;
    if ( $keepGoing == 0 )
    {
        print "$now_string: Got termination signal..\n";
        return -2;
    }


    if ($configValues{localHostNameAsCEHostName} eq "yes") {
	$hash{server} = $hostName.".".$domainName;
    }

    my $cmd = "";
    my $status = -1; 
    my $resGridId="$hash{server}:$hash{queue}";
    $cmd = "/opt/glite/libexec/glite_dgas_atmClient";
    $cmd .= " --jobid $hash{server}:$hash{lrmsId}_$hash{start}";
    $cmd .= " --resgridid $resGridId";
    $cmd .= " CPU_TIME=$hash{cput} WALL_TIME=$hash{walltime} PMEM=$hash{mem} VMEM=$hash{vmem}";
    #$cmd .= " \"QUEUE=$hash{queue}\" \"USER=$hash{user}\" \"LRMSID=$hash{lrmsId}\"";
    if (exists($hash{queue})) {
	$cmd .= " \"QUEUE=$hash{queue}\"";
    }
    if (exists($hash{user})) {
	$cmd .= " \"USER=$hash{user}\"";
    }
    if (exists($hash{lrmsId})) {
	$cmd .= " \"LRMSID=$hash{lrmsId}\"";
    }
    #$cmd .= " \"PROCESSORS=$hash{processors}\" \"URCREATION=$now_string\"";
    if (exists($hash{processors})) {
	$cmd .= " \"PROCESSORS=$hash{processors}\"";
    }
    $cmd .= " \"URCREATION=$now_string\"";
    #$cmd .= " \"group=$hash{group}\" \"jobName=$hash{jobName}\"";
    if (exists($hash{group})) {
	$cmd .= " \"group=$hash{group}\"";
    }
    if (exists($hash{jobName})) {
	$cmd .= " \"jobName=$hash{jobName}\"";
    }
    #$cmd .= " \"start=$hash{start}\" \"end=$hash{end}\"";
    if (exists($hash{start})) {
	$cmd .= " \"start=$hash{start}\"";
    }
    if (exists($hash{end})) {
	$cmd .= " \"end=$hash{end}\"";
    }
    #$cmd .= " \"ctime=$hash{ctime}\" \"qtime=$hash{qtime}\"";
    if (exists($hash{ctime})) {
	$cmd .= " \"ctime=$hash{ctime}\"";
    }
    if (exists($hash{qtime})) {
	$cmd .= " \"qtime=$hash{qtime}\"";
    }
    #$cmd .= " \"etime=$hash{etime}\" \"exitStatus=$hash{exitStatus}\"";
    if (exists($hash{etime})) {
	$cmd .= " \"etime=$hash{etime}\"";
    }
    if (exists($hash{exitStatus})) {
	$cmd .= " \"exitStatus=$hash{exitStatus}\"";
    }

    # we don't have this from the LRMS log, but from the LDIF file(s):
    ##$cmd .= " \"si2k=$hash{si2k}\" \"sf2k=$hash{sf2k}\"";
    #if (exists($hash{si2k})) {
	#$cmd .= " \"si2k=$hash{si2k}\"";
    #}
    #if (exists($hash{sf2k})) {
	#$cmd .= " \"sf2k=$hash{sf2k}\"";
    #}

    # If we found an LDIF file ($ldifModTimestamp != 0)
    # and this job is more recent than the file, then we trust the
    # SpecInt and SpecFloat values taken from the file:
    # (IMPORTANT: for now we DO NOT check the age of the job and ALWAYS
    #  accept the LDIF attributes!)
    if (exists($hash{start}) && ($ldifModTimestamp != 0) 
	#&& ($hash{start} >= $ldifModTimestamp) 
	) {
	if (exists($glueAttributes{"GlueHostApplicationSoftwareRunTimeEnvironment: SI00MeanPerCPU"})
	    && $glueAttributes{"GlueHostApplicationSoftwareRunTimeEnvironment: SI00MeanPerCPU"} ne "") {
	    my $si2k = $glueAttributes{"GlueHostApplicationSoftwareRunTimeEnvironment: SI00MeanPerCPU"};
	    $cmd .= " \"si2k=$si2k\"";
	    print "(note: Taking mean SI00 value per CPU, instead of the normal benchmark)\n";
	}
	elsif (exists($glueAttributes{GlueHostBenchmarkSI00})
	       && $glueAttributes{GlueHostBenchmarkSI00} ne "") {
	    $cmd .= " \"si2k=$glueAttributes{GlueHostBenchmarkSI00}\"";
	}

	if (exists($glueAttributes{"GlueHostApplicationSoftwareRunTimeEnvironment: SF00MeanPerCPU"})
	    && $glueAttributes{"GlueHostApplicationSoftwareRunTimeEnvironment: SF00MeanPerCPU"} ne "") {
	    my $sf2k = $glueAttributes{"GlueHostApplicationSoftwareRunTimeEnvironment: SF00MeanPerCPU"};
	    $cmd .= " \"sf2k=$sf2k\"";
	    print "(note: Taking mean SF00 value per CPU, instead of the normal benchmark)\n";
	}
	elsif (exists($glueAttributes{GlueHostBenchmarkSF00})
	       && $glueAttributes{GlueHostBenchmarkSF00} ne "") {
	    $cmd .= " \"sf2k=$glueAttributes{GlueHostBenchmarkSF00}\"";
	}
    }

    $cmd .= " \"tz=$timeZone\" \"accountingProcedure=outOfBand\"";
    #$cmd .= " \"fqan=/$hash{userVO}/Role=NULL/Capability=NULL\""; #!!!!!!
    if (exists($hash{userVO})) {
	$cmd .= " \"fqan=/$hash{userVO}/Role=NULL/Capability=NULL\"";
    }
    $cmd .= " \"ceHostName=$hash{server}\"";
    if (exists($hash{execHost})) {
	$cmd .= " \"execHost=$hash{execHost}\"";
    }
    if ( $configValues{siteName} ne "" )
    { 
            $cmd .= " \"SiteName=$configValues{siteName}\"";
    }
    $cmd .= " --usrcert \"$hash{user}:$hash{server}\"";
    $cmd .= " -3";
    $status = system("$cmd");
    print "$now_string: Running: $cmd\n";
    if ($? == -1) {
           print "failed to execute: $!\n";
    }
    elsif ($? & 127) {
           $status = ($? & 127);
    }
    else {
           $status = $? >> 8;
    }
    print "EXITSTATUS=$status\n";

    if ( $status == 64 || ($status>=67 && $status<=69) ) {
	# from hlr_prot_errcode.h:
        #define ATM_E_TRANS 		"64"
        #define ATM_E_TRANS_A 		"67"
        #define ATM_E_TRANS_B 		"68"
        #define ATM_E_TRANS_C 		"69"
	print "Attention: The record cannot be inserted into the HLRs transaction queue. Terminating, job will be retried next time (if this conditions persists, there might be problems on the HLR (database down, record with same jobID already in queue cannot be processed, etc).\n";
    }

    if ( $status == 65 || ($status>=70 && $status<=73) ) {
	# from hlr_prot_errcode.h:
        #define ATM_E_DUPLICATED 	"65"
        #define ATM_E_DUPLICATED_A 	"70"
        #define ATM_E_DUPLICATED_B 	"71"
        #define ATM_E_DUPLICATED_C 	"72"
        #define ATM_E_DUPLICATED_D 	"73"
	print "The record is already accounted on the HLR.\n";
	return 0;
    }

    print "\n";

    return $status;
}


##-------> lock  subroutines <---------##

sub putLock
{
    my $lockName = $_[0];
    open(IN,  "< $lockName") && return 1;
    close(IN);
    open(OUT, "> $lockName") || return 2;
    print OUT  $$;    ## writes pid   
    close(OUT);
    return 0;
}

sub delLock
{
    my $lockName = $_[0];
    open(IN,  "< $lockName") || return 1;
    close(IN);
    my $status = system("rm -f $lockName");
    return $status;
}

##-------> sig handlers subroutines <---------##

sub putBuffer
{
    my $buffName = $_[0];
    if ($keepGoing == 1) {
	# this is done only if no SIGINT was received!
	open(OUT, "> $buffName") || return 2;
	print OUT  "$_[1]:$_[2]";    ## writes lastId
	close(OUT);
    }
    else {
	print "SIGINT received, buffer will not be written!\n";
    }
    return 0;
}

sub readBuffer
{
	my $buffname = $_[0];
	open(IN, "< $buffname") || return 2;
	my $line;
	my $file;
	while ( <IN> )
	{
		($line,$file) = split(':');
		chomp($line);
		chomp($file);
		$_[1] = $line;
		$_[2] = $file;
	}	
	close(IN);
	return 0;
}

sub sigHUP_handler {
        print "got SIGHUP\n";
	$keepGoing = 1;
}

sub sigINT_handler {
        print "got SIGINT\n";
        $keepGoing = 0;
}


## ------ get GLUE attributes From LDIF file --------- ##
sub getGLUEAttributesFromLdif {
    my $modTStamp = 0; # returns 0 if no valid attributes found!
                       # otherwise returns the timestamp of the last
                       # modification of the file!

    # to be stored in %glueAttributes
    %glueAttributes = (); # first empty everything

    my @ldifFiles = split(/,/, $ldifDefaultFiles);
    unshift(@ldifFiles, $glueLdifFile); # first to try!

    my @keys = split(/,/, $keyList);

    if (!@keys) {
	print "".localtime().": Warning: No GLUE attributes will be added to usage records (reason: no keyList in configuration file)!\n";
	return 0; # no keys -> no benchmarks
    }

    while (@ldifFiles) {
	my $file = shift(@ldifFiles);
	print "".localtime().": Trying to get GLUE benchmarks from LDIF file: $file\n";

	if (!open(GLUEFILE, "< $file")) {
	    print "Warning: could not open the LDIF file ... skipping!\n";
	    next;
	}

	my $foundSomething = 0;
	my $line;
	while ($line = <GLUEFILE>) {
	    my $key;
	    foreach $key (@keys) {
		if ( ($line =~ /^$key:\s?(.*)$/ )
		     || ($line =~ /^$key=(.*)$/ ) 
		     || ($line =~ /^${key}_(.*)$/ ) ) {
		    # accept stuff like "GlueHostBenchmarkSI00: 955" and
		    # "GlueHostApplicationSoftwareRunTimeEnvironment: SI00MeanPerCPU=955"
		    # "GlueHostApplicationSoftwareRunTimeEnvironment: SI00MeanPerCPU_955"
		    $glueAttributes{$key} = $1;
		    print "found: $key=$1; ";
		    $foundSomething = 1;
		}
	    }
	}

	close(GLUEFILE);

	if ($foundSomething) {
	    print "\n";
	    # get timestamp of file:
	    $modTStamp = (stat($file))[9]; #mtime in sec. after epoch

	    last; # don't check the other files (if any)!
	}

    }

    my $thisTStamp = time();

    if ($modTStamp == 0) {
	print "".localtime().": Warning: No GLUE attributes will be added to usage records (reason: no valid entries in LDIF file(s))!\n";
	return 0;
    }
    else {
	if($modTStamp > $thisTStamp-$DEF_LDIF_VALIDITY) {
	    # accept it for at least a day, even if more recently modified
	    $modTStamp = $thisTStamp-$DEF_LDIF_VALIDITY;
	}
	# we disabled the controll above!
	#print "GLUE attributes will be accepted for all jobs with LRMS creation time after $modTStamp\n";
    }



    return $modTStamp;
}
