#!/usr/bin/perl -w

#Gianduia, DGAS job monitor and Usage Records collector.
use strict;
use POSIX;
use Text::ParseWords;

my $sigset = POSIX::SigSet ->new();
my $actionHUP=POSIX::SigAction->new("sigHUP_handler",$sigset,&POSIX::SA_NODEFER);
my $actionInt=POSIX::SigAction->new("sigINT_handler",$sigset,&POSIX::SA_NODEFER);
POSIX::sigaction(&POSIX::SIGHUP, $actionHUP);
POSIX::sigaction(&POSIX::SIGINT, $actionInt);
POSIX::sigaction(&POSIX::SIGTERM, $actionInt);
my $configFilePath ="/opt/glite/etc/glite-dgas-OOBCollector.conf"; 
my %configValues = (
		    collectorAcctLogDir     =>  "",
		    lockFileName      => "/opt/glite/var/dgasOOB.lock",
		    bufferFileName      => "/opt/glite/var/dgasOOBuffer",
		    mainPollInterval  => "5",
		    queuePollInterval =>"25",
		    timeInterval      =>"5",
		    jobPerTimeInterval =>"10",
		    fileTemplate    =>".raw",
		    siteName    => "",
		    );

my %hash = (
             server   =>  "server",
             queue    =>  "queue",
             lrmsType =>  "lsf",
             globusResContact =>  "GLOBUS_RES_CONTACT",
             cput     =>  "cput",
             hlrLoc   =>  "HLR_LOCATION",
             jobId    =>  "WMS_JOBID",
             jobIdEdg =>  "EDG_WL_JOBID",
             user     =>  "PBS_O_LOGNAME",
             mem      =>  "mem",
             vmem     =>  "vmem",
             walltime =>  "walltime",
             lrmsId   =>  "lrmsId",           
             jobName  =>  "jobName",
             group    =>  "group", 
             ctime    =>  "ctime",
             qtime    =>  "qtime",
             etime    =>  "etime",
             start    =>  "start", 
             end      =>  "end",
             exitStatus      =>  "exitStatus", 
             execHost    =>  "execHost",
             processors => "processors",
             si2k => "GlueHostBenchmarkSI00",
             sf2k => "GlueHostBenchmarkSF00",
             tz => "timeZone",
             userVO => "",
             );

#Parse configuration file
if(exists $ARGV[0]) 
{
    $configFilePath = $ARGV[0];
}	
&parseConf($configFilePath); 

## Global variable initialization
my $UR;
my $list;

my $collectorAcctLogDir    = $configValues{collectorAcctLogDir};
my $timeInterval     = $configValues{timeInterval};
my $jobPerTimeInterval     = $configValues{jobPerTimeInterval};

my $lockFileName = $configValues{lockFileName};

my $T2 = $configValues{queuePollInterval};
my $T1 = $configValues{mainPollInterval};
my $garbageCleanInterval = 10*$T2;

my $timeZone = `date +%z`;
chomp($timeZone);
my $domainName = `hostname -d`;
chomp($domainName);
my $hostName = `hostname`;
chomp($hostName);
my $keepGoing = 1;
my $start1 = time;   ##  time init
my $start2 = time;   ##  time init
my @garbageQueue = ();

my $fileTemplate = $configValues{fileTemplate};
if ($ARGV[1] ne "")
{
	$fileTemplate = $ARGV[1];
}

#put lock
if ( &putLock($lockFileName) != 0 )
{
    die "Fatal Error: Couldn't open lock file! in $lockFileName\n";
}
else
{
	print "Lock file succesfully created.\n";
}

#This script must be exexuted in the cron.
my $startJob = "";
my $startFile = "";
my $lastJob = "";
my $lastFile = "";
&readBuffer($configValues{bufferFileName}, $startJob, $startFile);
print "First job to analyse: $startFile:$startJob\n";
&checkPbsLog($startJob,$startFile,$lastFile,$lastJob);    
print "Last job analysed: $lastFile:$lastJob\n";
if ($lastJob ne "" )
{
	&putBuffer($configValues{bufferFileName}, $lastJob, $lastFile);
}


print "Exiting...\n";
if ( &delLock($lockFileName) != 0 )
{	
    print "Error removing lock file.\n";
}
else
{
    print "Lock file removed.\n";
}
print "Exit.\n";


###-----------------------END-----------------------------------------###

sub parseConf 
{
    my $fconf = $_[0];
    open(FILE, "$fconf") || die "file not open \n";
    while(<FILE>)
    {
	if(/^collectorAcctLogDir\s*=\s*\"(.*)\"$/){$configValues{collectorAcctLogDir}=$1;}
	if(/^fileTemplate\s*=\s*\"(.*)\"$/){$configValues{fileTemplate}=$1;}
	if(/^oobLockFileName\s*=\s*\"(.*)\"$/){$configValues{lockFileName}=$1;}
	if(/^bufferFileName\s*=\s*\"(.*)\"$/){$configValues{bufferFileName}=$1;}
	if(/^mainPollInterval\s*=\s*\"(.*)\"$/){$configValues{mainPollInterval}=$1;}
	if(/^queuePollInterval\s*=\s*\"(.*)\"$/){$configValues{queuePollInterval}=$1;}
	if(/^siteName\s*=\s*\"(.*)\"$/){$configValues{siteName}=$1;}
    }
    close(FILE);
}

sub checkPbsLog#FIXME
{
        my $startJob = $_[0]; #jobNumber to start with.
        my $startFile = $_[1]; #jobNumber to start with.
	my $lastFile= $_[2]; #last file analysed on previus run
        $collectorAcctLogDir= $configValues{collectorAcctLogDir};
	my @dirs;
	opendir(DIR,$collectorAcctLogDir ) || die "can't opendir $collectorAcctLogDir: $!";
	while( defined(my $file = readdir(DIR)) )
	{
		if ( $file =~ /$fileTemplate$/ )
		{
			push @dirs, $file;
		}
	}
	closedir DIR;
	my $goOn = 1;
	my $firstJobId = 1;
	my $firstFileOpened = 0;
	print "Scanning logdir $collectorAcctLogDir until file:$startFile\n";
	while ( $goOn && $keepGoing && @dirs && !$firstFileOpened )
	{
		my $file = pop @dirs; #going backward
		next if (($file eq ".") || ($file eq ".."));
		if ( $file eq $startFile )
		{
		    ##this file has already been fully processed
		    ##so we can break here.
		    #print "Found $startFile in dir no need to go further!\n";
		    #return 0;
		    print "Found $startFile in dir; last file to process!\n";
		    $firstFileOpened = 1;
		}
		print "Processing: $file \n";

		open ( LOGFILE, "/usr/bin/tac $collectorAcctLogDir/$file |" ) or return 1;
		print "Parsing file from end to $startJob in $collectorAcctLogDir/$file\n";
		while ( <LOGFILE> )
		{
			if ( $keepGoing == 0 )
			{
				print "Exiting from PBS loop...\n";
				close LOGFILE;
				return 1;
			}
			my @ARRAY = split(" " , $_ );
			my $targetJobId = $ARRAY[1];
			if ( ( $startJob eq $targetJobId ) && ($startJob ne "") )
			{
			    #Got the right line!
			    print "Found $targetJobId in LOG!\n";
			    close LOGFILE;
			    return 0;
			}	
			else
			{
				if ( $firstJobId == 1 )
				{
					$_[2] = $file;
					$_[3] = $targetJobId;
					$firstJobId = 0;
				}
				print "$targetJobId:";
				&parseUR_Collector($_);
				print "processing $hash{jobName}\n";
				&callAtmClient();
			}
		}
	}
}

sub parseUR_Collector
{
	my $URString = $_[0];
	my @URArray = split ( ' ' , $URString );
	$hash{lrmsId}=$URArray[1];
	$hash{user}=$URArray[2];
	$hash{queue}=$URArray[3];
	$hash{qtime}=$URArray[4];
	$hash{start}=$URArray[5];
	$hash{end}=$URArray[6];
	$hash{etime}=$URArray[7]+$URArray[4];
	$hash{walltime}=$URArray[8];
	$hash{cput}=$URArray[9]+$URArray[10];
	$hash{mem}=$URArray[11]*1024;#redEye uses MB DGAS uses KB
	$hash{vmem}=$URArray[12]*1024;#redEye uses MB DGAS uses KB
	$hash{server}=$URArray[13];
	$hash{execHost}=$URArray[14];
	$hash{si2k}=$URArray[15]*1000;#redEye uses KSI2k, DGAS uses si2k 
	$hash{jobName}="$hash{user}:$hash{server}_$hash{lrmsId}:$hash{execHost}";#redEye uses KSI2k, DGAS uses si2k 
	&detectVOfromqueue();
}

sub detectVOfromuid
{
    $_=$_[0];
    if(/(.*)\d\d\d/)
    {
	$hash{userVO}=$1;
    }
    else
    {
	$hash{userVO}="";
    }
}

sub detectVOfromqueue
{
    $_=$_[0];
    $hash{userVO}=$hash{queue};
}

sub callAtmClient 
{
    my $now_string = localtime;
    if ( $keepGoing == 0 )
    {
        print "$now_string: Got termination signal..\n";
        return -2;
    }
    my $cmd = "";
    my $status = -1; 
    my $resGridId="$hostName:$hash{queue}";
    $cmd = "/opt/glite/libexec/glite_dgas_atmLocal";
    $cmd .= " --jobid $hash{server}:$hash{lrmsId}_$hash{start}";
    $cmd .= " --resgridid $resGridId";
    $cmd .= " CPU_TIME=$hash{cput} WALL_TIME=$hash{walltime} PMEM=$hash{mem} VMEM=$hash{vmem}";
    $cmd .= " \"QUEUE=$hash{queue}\" \"USER=$hash{user}\" \"LRMSID=$hash{lrmsId}\"";
    $cmd .= " \"PROCESSORS=$hash{processors}\" \"URCREATION=$now_string\"";
    $cmd .= " \"group=$hash{group}\" \"jobName=$hash{execHost}:$hash{jobName}\"";
    $cmd .= " \"start=$hash{start}\" \"end=$hash{end}\"";
    $cmd .= " \"ctime=$hash{ctime}\" \"qtime=$hash{qtime}\"";
    $cmd .= " \"etime=$hash{etime}\" \"exitStatus=$hash{exitStatus}\"";
    $cmd .= " \"si2k=$hash{si2k}\" \"sf2k=$hash{sf2k}\" \"execHost=$hash{execHost}\"";
    $cmd .= " \"tz=$timeZone\" \"accountingProcedure=outOfBand\"";
    $cmd .= " \"fqan=/$hash{userVO}/Role=NULL/Capability=NULL\""; #!!!!!!
    $cmd .= " \"userVo=$hash{userVO}\""; #!!!!!!
    $cmd .= " \"ceHostName=$hash{server}\"";
    if ( $configValues{siteName} ne "" )
    { 
	    $cmd .= " \"SiteName=$configValues{siteName}\""; 
    }
    $cmd .= " --usrcert \"$hash{user}:$hash{server}\"";
    $cmd .= " -3 >/dev/null";
    $status = system("$cmd");
    print $cmd;
#    print "$now_string: Running: $cmd\n";
    if ($? == -1) {
           print "failed to execute: $!\n";
    }
    elsif ($? & 127) {
           $status = ($? & 127);
    }
    else {
           $status = $? >> 8;
    }
    print "\nEXITSTATUS=$status\n\n";
    return $status;
}


##-------> lock  subroutines <---------##

sub putLock
{
    my $lockName = $_[0];
    open(IN,  "< $lockName") && return 1;
    close(IN);
    open(OUT, "> $lockName") || return 2;
    print OUT  $$;    ## writes pid   
    close(OUT);
    return 0;
}

sub delLock
{
    my $lockName = $_[0];
    open(IN,  "< $lockName") || return 1;
    close(IN);
    my $status = system("rm -f $lockName");
    return $status;
}

##-------> sig handlers subroutines <---------##

sub putBuffer
{
    my $buffName = $_[0];
    if ($keepGoing == 1) {
	# this is done only if no SIGINT was received!
	open(OUT, "> $buffName") || return 2;
	print OUT  "$_[1]:$_[2]";    ## writes lastId
	close(OUT);
    }   
    return 0;
}

sub readBuffer
{
	my $buffname = $_[0];
	open(IN, "< $buffname") || return 2;
	my $line;
	my $file;
	while ( <IN> )
	{
		($line,$file) = split(':');
		$_[1] = $line;
		$_[2] = $file;
	}	
	close(IN);
	return 0;
}

sub sigHUP_handler {
        print "got SIGHUP\n";
	$keepGoing = 1;
}

sub sigINT_handler {
        print "got SIGINT\n";
        $keepGoing = 0;
}
